#!/usr/bin/env python
# -*- coding: utf-8 -*-
# import scipy as sp
import sys
import numpy as np
from six.moves import xrange
from pyssp.util import read_signal, get_frame


WINSIZE = 2048


class LTSD():
    def __init__(self, winsize, window, order, e0=50, e1=80, lambda0=30, lambda1=12):
        self._winsize = winsize
        self._window = window
        self._order = order
        self._amplitude = {}
        # calcurate parameters from datasets
        self._e0 = e0
        self._e1 = e1
        self._lambda0 = lambda0
        self._lambda1 = lambda1

    def _get_amplitude(self, signal, l):
        if l in self._amplitude:
            return self._amplitude[l]
        else:
            frame = get_frame(signal, self._winsize, l)
            spec = np.fft.rfft(frame * self._window)
            amp = np.absolute(spec)
            # amp = sp.absolute(sp.fft(frame * self._window))
            self._amplitude[l] = amp
            return amp

    def _compute_noise_avgspectrum(self, nsignal):
        windownum = int(len(nsignal)/(self._winsize/2) - 1)
        avgamp = np.zeros(int(self._winsize/2) + 1)
        for l in xrange(windownum):
            frame = get_frame(nsignal, self._winsize, l)
            avgamp += np.absolute(np.fft.rfft(frame * self._window))
        return avgamp/float(windownum)

    def _is_signal(self, signal, ltsd, l):
        e = self._calc_power(signal, l)
        print(e, ltsd)
        if e < self._e0:
            if ltsd > self._lambda0:
                return True
            else:
                return False
        elif e > self._e1:
            if ltsd > self._lambda1:
                return True
            else:
                return False
        else:
            lamb = ((self._lambda0 - self._lambda1) / (self._e0 - self._e1)) * e +\
                self._lambda0 - (self._lambda0 - self._lambda1) / (1 - self._e1/self._e0)
            print(lamb)
            if ltsd > lamb:
                return True
            else:
                return False

    def _calc_power(self, signal, l):
        amp = self._get_amplitude(signal, l)
        # print(np.argmax(amp), amp[np.argmax(amp)], np.average(amp))
        # print(np.average(amp))
        # print(10*np.log10((np.sum(amp)/amp.shape[0])**2))
        # avg = 10.0*np.log10(np.sum(amp**2)/float(amp.shape[0]))
        avg = 10*np.log10((np.average(amp) / (10e-7 * 20)) ** 2)
        return avg

    def compute_without_noise(self, signal, size):
        self._windownum = int(len(signal)/(self._winsize/2) - 1)
        # Calcurate the average noise spectrum amplitude based　on first 'size' bytes in the head part of input signal.
        self._avgnoise = self._compute_noise_avgspectrum(signal[0:size])**2
        return self._compute(signal)

    def compute_with_noise(self, signal, noise):
        self._windownum = len(signal)/(self._winsize/2)-1
        self._avgnoise = self._compute_noise_avgspectrum(noise)**2
        return self._compute(signal)

    def _compute(self, signal):
        ltsds = np.zeros(self._windownum)
        prev = 0
        pair = None
        result = []
        for l in xrange(self._windownum):
            ltsd = self._ltsd(signal, l)
            ltsds[l] = ltsd
            x = self._is_signal(signal, ltsd, l)
            if x:  # signal
                if prev == 0:  # start signal segment
                    pair = [l]
                prev = 1
            else:  # noise
                if prev == 1:  # end signal segment
                    pair.append(l-1)
                    result.append(pair)
                    pair = None
                prev = 0
        return result, ltsds

    def _ltse(self, signal, l):
        maxamp = np.zeros(int(self._winsize/2) + 1)
        for idx in range(l-self._order, l+self._order+1):
            amp = self._get_amplitude(signal, idx)
            maxamp = np.maximum(maxamp, amp)
        return maxamp

    def _ltsd(self, signal, l):
        if l < self._order or l+self._order >= self._windownum:
            return 0
        return 10.0 * np.log10(np.average(self._ltse(signal, l)**2/self._avgnoise))


class AdaptiveLTSD(LTSD):
    def __init__(self, winsize, window, order, ratio=0.95, e0=200, e1=300, lambda0=40, lambda1=50):
        self._ratio = ratio
        LTSD.__init__(self, winsize, window, order, e0, e1, lambda0, lambda1)

    def _update_noise_spectrum(self, signal, l):
        avgamp = np.zeros(self._winsize)
        for idx in range(l - self._order, l + self._order + 1):
            avgamp += self._get_amplitude(signal, idx)
        avgamp = avgamp / float(self._order*2 + 1)
        self._avgnoise = self._avgnoise * self._ratio + (avgamp**2)*(1.0-self._ratio)

    def _compute(self, signal):
        ltsds = np.zeros(self._windownum)
        prev = 0
        pair = None
        result = []
        for l in xrange(self._windownum):
            ltsd = self._ltsd(signal, l)
            ltsds[l] = ltsd
            x = self._is_signal(signal, ltsd, l)
            if x:  # signal
                if prev == 0:  # start signal segment
                    pair = [l]
                prev = 1
            else:  # noise
                if prev == 1:  # end signal segment
                    pair.append(l-1)
                    result.append(pair)
                    self._update_noise_spectrum(signal, l)
                    pair = None
                prev = 0
        return result, ltsds


if __name__ == "__main__":
    signal, params = read_signal(sys.argv[1], WINSIZE)
    window = np.hanning(WINSIZE)

    ltsd = LTSD(WINSIZE, window, 5)
    res, ltsds = ltsd.compute_without_noise(signal, WINSIZE * int(params[2] / float(WINSIZE) / 3.0))
    print(ltsds)
    # import matplotlib.pyplot as plt
    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.plot(ltsds)
    # plt.show()
    print(res)



# -*- coding: utf-8 -*-
import wave
import scipy as sp
from util import read_signal, get_frame, add_signal, write_signal, compute_avgpowerspectrum
from pyssp.voice_enhancement import JointMap
from six.move import xrange
from .MinimumStatistics import MinimumStatistics

if __name__ == "__main__":
    WINSIZE = 512
    samplingrate = 16000
    chunk = 1024
    soundfile = 'tools/sound/F2AKFU.SD.E03.16k.2.wav'
    soundfile = 'tools/sound/b01.wav'
    noisefile = 'tools/sound/air_vent.16k.wav'
    sound = wave.open(soundfile, 'rb')
    noise = wave.open(noisefile, 'rb')

    synthfile = 'tools/sound/noisy.wav'
    synth = wave.open(synthfile, 'wb')
    synth.setnchannels(1)
    synth.setsampwidth(2)
    synth.setframerate(samplingrate)

    remain = sound.getnframes()

    while remain > 0:
        s = min(chunk, remain)
        # read frames
        data_sound = sound.readframes(s)
        data_noise = noise.readframes(s)
        # convert
        ary_sound = sp.fromstring(data_sound, sp.int16)
        ary_noise = sp.fromstring(data_noise, sp.int16)

        int32_ary_sound = sp.int32(ary_sound)
        int32_ary_noise = sp.int32(ary_noise)
        ary2 = sp.int16(int32_ary_sound + int32_ary_noise)
        data2 = ary2.tostring()
        synth.writeframes(data2)
        remain = remain - s
    sound.close()
    noise.close()
    synth.close()

    infile = 'tools/sound/noisy.wav'
    signal, params = read_signal(infile, WINSIZE)
    nf = len(signal) / (WINSIZE / 2) - 1
    sig_out = sp.zeros(len(signal), sp.float32)
    window = sp.hanning(WINSIZE)

    ms = MinimumStatistics(WINSIZE, window, params[2])
    NP_lambda = compute_avgpowerspectrum(signal[0:WINSIZE * int(params[2] / float(WINSIZE) / 3.0)],
                                         WINSIZE, window)
    ms.init_noise_profile(NP_lambda)
    ss = JointMap(WINSIZE, window)
    for no in xrange(nf):
        frame = get_frame(signal, WINSIZE, no)
        n_pow = ms.compute(frame, no)
        res = ss.compute_by_noise_pow(frame, n_pow)
        add_signal(sig_out, res, WINSIZE, no)

    ms.show_debug_result()
    write_signal("tools/sound/noise_reduction.wav", params, sig_out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import scipy.special as spc
from pyssp.util import sigmoid
from six.moves import xrange


class SpectralSubtruction:
    def __init__(self, winsize, window, coefficient=5.0, ratio=1.0):
        self._window = window
        self._coefficient = coefficient
        self._ratio = ratio

    def compute(self, signal, noise):
        n_spec = np.fft.fftpack.fft(noise * self._window)
        n_pow = np.absolute(n_spec) ** 2.0
        return self.compute_by_noise_pow(signal, n_pow)

    def compute_by_noise_pow(self, signal, n_pow):
        s_spec = np.fft.fftpack.fft(signal * self._window)
        s_amp = np.absolute(s_spec)
        s_phase = np.angle(s_spec)
        amp = s_amp ** 2.0 - n_pow * self._coefficient
        amp = np.maximum(amp, 0.0)
        amp = np.sqrt(amp)
        amp = self._ratio * amp + (1.0 - self._ratio) * s_amp
        spec = amp * np.exp(s_phase * 1j)
        return np.real(np.fft.fftpack.ifft(spec))


class SpectrumReconstruction:
    def __init__(self, winsize, window, constant=0.001, ratio=1.0, alpha=0.99):
        self._window = window
        self._G = np.zeros(winsize, np.float64)
        self._prevGamma = np.zeros(winsize, np.float64)
        self._alpha = alpha
        self._prevAmp = np.zeros(winsize, np.float64)
        self._ratio = ratio
        self._constant = constant

    def _sigmoid(self, gain):
        for i in xrange(len(gain)):
            gain[i] = sigmoid(gain[1], 1, 2, self._gain)

    def compute(self, signal, noise):
        n_spec = np.fft.fftpack.fft(noise * self._window)
        n_pow = np.absolute(n_spec) ** 2.0
        return self.compute_by_noise_pow(signal, n_pow)

    def _calc_aposteriori_snr(self, s_amp, n_pow):
        return s_amp ** 2.0 / n_pow

    def _calc_apriori_snr(self, gamma):
        return self._alpha * self._G ** 2.0 * self._prevGamma +\
            (1.0 - self._alpha) * np.maximum(gamma - 1.0, 0.0)  # a priori s/n ratio

    def _calc_apriori_snr2(self, gamma, n_pow):
        return self._alpha * (self._prevAmp ** 2.0 / n_pow) +\
            (1.0 - self._alpha) * np.maximum(gamma - 1.0, 0.0)  # a priori s/n ratio


class MMSE_STSA(SpectrumReconstruction):
    def __init__(self, winsize, window, constant=0.001, ratio=1.0, alpha=0.99):
        self._gamma15 = spc.gamma(1.5)
        super(MMSE_STSA, self).__init__(winsize, window, constant=constant, ratio=ratio, alpha=alpha)

    def compute_by_noise_pow(self, signal, n_pow):
        s_spec = np.fft.fftpack.fft(signal * self._window)
        s_amp = np.absolute(s_spec)
        s_phase = np.angle(s_spec)
        gamma = self._calc_aposteriori_snr(s_amp, n_pow)
        xi = self._calc_apriori_snr(gamma)
        self._prevGamma = gamma
        nu = gamma * xi / (1.0 + xi)
        self._G = (self._gamma15 * np.sqrt(nu) / gamma) * np.exp(-nu / 2.0) *\
                  ((1.0 + nu) * spc.i0(nu / 2.0) + nu * spc.i1(nu / 2.0))
        idx = np.less(s_amp ** 2.0, n_pow)
        self._G[idx] = self._constant
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = xi[idx] / (xi[idx] + 1.0)
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = self._constant
        self._G = np.maximum(self._G, 0.0)
        amp = self._G * s_amp
        amp = np.maximum(amp, 0.0)
        amp2 = self._ratio * amp + (1.0 - self._ratio) * s_amp
        self._prevAmp = amp
        spec = amp2 * np.exp(s_phase * 1j)
        return np.real(np.fft.fftpack.ifft(spec))


class MMSE_LogSTSA(SpectrumReconstruction):
    def __init__(self, winsize, window, constant=0.001, ratio=1.0, alpha=0.99):
        self._gamma15 = spc.gamma(1.5)
        super(MMSE_LogSTSA, self).__init__(winsize, window, constant=constant, ratio=ratio, alpha=alpha)

    def compute_by_noise_pow(self, signal, n_pow):
        s_spec = np.fft.fftpack.fft(signal * self._window)
        s_amp = np.absolute(s_spec)
        s_phase = np.angle(s_spec)
        gamma = self._calc_aposteriori_snr(s_amp, n_pow)
        xi = self._calc_apriori_snr(gamma)
        # xi = self._calc_apriori_snr2(gamma,n_pow)
        self._prevGamma = gamma
        nu = gamma * xi / (1.0 + xi)
        self._G = xi / (1.0 + xi) * np.exp(0.5 * spc.exp1(nu))
        idx = np.less(s_amp ** 2.0, n_pow)
        self._G[idx] = self._constant
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = xi[idx] / (xi[idx] + 1.0)
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = self._constant
        self._G = np.maximum(self._G, 0.0)
        amp = self._G * s_amp
        amp = np.maximum(amp, 0.0)
        amp2 = self._ratio * amp + (1.0 - self._ratio) * s_amp
        self._prevAmp = amp
        spec = amp2 * np.exp(s_phase * 1j)
        return np.real(np.fft.fftpack.ifft(spec))


class JointMap(SpectrumReconstruction):
    def __init__(self, winsize, window, constant=0.001, ratio=1.0, alpha=0.99, mu=1.74, tau=0.126):
        self._mu = mu
        self._tau = tau
        super(JointMap, self).__init__(winsize, window, constant=constant, ratio=ratio, alpha=alpha)

    def compute_by_noise_pow(self, signal, n_pow):
        s_spec = np.fft.fftpack.fft(signal * self._window)
        s_amp = np.absolute(s_spec)
        s_phase = np.angle(s_spec)
        gamma = self._calc_aposteriori_snr(s_amp, n_pow)
        # xi = self._calc_apriori_snr2(gamma,n_pow)
        xi = self._calc_apriori_snr(gamma)
        self._prevGamma = gamma
        u = 0.5 - self._mu / (4.0 * np.sqrt(gamma * xi))
        self._G = u + np.sqrt(u ** 2.0 + self._tau / (gamma * 2.0))
        idx = np.less(s_amp ** 2.0, n_pow)
        self._G[idx] = self._constant
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = xi[idx] / (xi[idx] + 1.0)
        idx = np.isnan(self._G) + np.isinf(self._G)
        self._G[idx] = self._constant
        self._G = np.maximum(self._G, 0.0)
        amp = self._G * s_amp
        amp = np.maximum(amp, 0.0)
        amp2 = self._ratio * amp + (1.0 - self._ratio) * s_amp
        self._prevAmp = amp
        spec = amp2 * np.exp(s_phase * 1j)
        return np.real(np.fft.fftpack.ifft(spec))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from six.move import xrange
from util import get_frame, read_signal


def itakura_saito_spectrum_distance(s, shat, winfunc):
    size = min(len(s), len(shat))
    window = winfunc(size)
    s = s[0:size]
    shat = shat[0:size]
    s_amp = np.absolute(np.fft.fftpack.fft(s * window))
    shat_amp = np.absolute(np.fft.fftpack.fft(shat * window))
    return np.mean(np.log(s_amp / shat_amp) + (shat_amp / s_amp) - 1.0)


def segmental_itakura_saito_spectrum_distance(s, shat, winsize, winfunc):
    size = min(len(s), len(shat))
    nf = size / (winsize / 2) - 1
    ret = []
    for no in xrange(nf):
        s_i = get_frame(s, winsize, no)
        shat_i = get_frame(shat, winsize, no)
        ret.append(itakura_saito_spectrum_distance(s_i, shat_i, winfunc))
    return ret


def log_spectrum_distance(s, shat, winfunc):
    size = min(len(s), len(shat))
    window = winfunc(size)
    s = s[0:size]
    shat = shat[0:size]
    s_amp = np.absolute(np.fft.fftpack.fft(s * window))
    shat_amp = np.absolute(np.fft.fftpack.fft(shat * window))
    return np.sqrt(np.mean((np.log10(s_amp / shat_amp) * 10.0) ** 2.0))


def segmental_log_spectrum_distance(s, shat, winsize, winfunc):
    size = min(len(s), len(shat))
    nf = size / (winsize / 2) - 1
    ret = []
    for no in xrange(nf):
        s_i = get_frame(s, winsize, no)
        shat_i = get_frame(shat, winsize, no)
        ret.append(log_spectrum_distance(s_i, shat_i, winfunc))
    return ret


if __name__ == "__main__":
    import sys
    winsize = int(sys.argv[1])
    s = read_signal(sys.argv[2], winsize)[0]
    shat = read_signal(sys.argv[3], winsize)[0]
    sissd = segmental_itakura_saito_spectrum_distance(s, shat, winsize, np.hanning)
    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(sissd)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#reference http://kaiseki-web.lhd.nifs.ac.jp/wiki/index.php/Python_%E3%81%AB%E3%82%88%E3%82%8B%E3%82%AA%E3%83%BC%E3%83%87%E3%82%A3%E3%82%AA%E5%87%A6%E7%90%86

import wave
import numpy as np
import scipy as sp
from six.moves import xrange, zip


def read_signal(filename, winsize):
    wf = wave.open(filename, 'rb')
    n = wf.getnframes()
    st = wf.readframes(n)
    params = ((wf.getnchannels(), wf.getsampwidth(),
               wf.getframerate(), wf.getnframes(),
               wf.getcomptype(), wf.getcompname()))
    siglen = ((int)(len(st) / 2 / winsize) + 1) * winsize
    signal = np.zeros(siglen, np.float32)
    signal[0:int(len(st) / 2)] = np.float32(np.fromstring(st, np.int16)) / 32767.0
    return [signal, params]


def get_frame(signal, winsize, no):
    shift = int(winsize / 2)
    start = int(no * shift)
    end = start + winsize
    return signal[start:end]


def add_signal(signal, frame, winsize, no):
    shift = int(winsize / 2)
    start = int(no * shift)
    end = start + winsize
    signal[start:end] = signal[start:end] + frame


def write_signal(filename, params, signal):
    wf = wave.open(filename, 'wb')
    wf.setparams(params)
    s = np.int16(signal * 32767.0).tostring()
    wf.writeframes(s)


def get_window(winsize, no):
    shift = int(winsize / 2)
    s = int(no * shift)
    return (s, s + winsize)


def separate_channels(signal):
    return signal[0::2], signal[1::2]


def uniting_channles(leftsignal, rightsignal):
    ret = []
    for i, j in zip(leftsignal, rightsignal):
        ret.append(i)
        ret.append(j)
    return np.array(ret, np.float32)


def compute_avgamplitude(signal, winsize, window):
    windownum = int(len(signal) / (winsize / 2)) - 1
    avgamp = np.zeros(winsize)
    for l in xrange(windownum):
        avgamp += np.absolute(sp.fft(get_frame(signal, winsize, l) * window))
    return avgamp / float(windownum)


def compute_avgpowerspectrum(signal, winsize, window):
    windownum = int(len(signal) / (winsize / 2)) - 1
    avgpow = np.zeros(winsize)
    for l in xrange(windownum):
        avgpow += np.absolute(sp.fft(get_frame(signal, winsize, l) * window))**2.0
    return avgpow / float(windownum)


def sigmoid(x, x0, k, a):
    y = k * 1 / (1 + np.exp(-a * (x - x0)))
    return y


def calc_kurtosis(samples):
    n = len(samples)
    avg = np.average(samples)
    moment2 = np.sum((samples - avg) ** 2) / n
    s_sd = np.sqrt(((n / (n - 1)) * moment2))
    k = ((n * (n + 1)) / ((n - 1) * (n - 2) * (n - 3))) * np.sum(((samples - avg) / s_sd) ** 4)
    return k - 3 * ((n - 1) ** 2) / ((n - 2) * (n - 3))



# -*- coding: utf-8 -*-
import numpy as np


class MMSEEstimator:
    # ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=6111268
    def __init__(self, winsize, window):
        self._winsize = winsize
        self._window = window
        self._alpha_ph1_mean = 0.9
        self._alpha_psd = 0.8
        self._q = 0.5
        self._prior_fact = self._q / (1 - self._q)
        self._xi_opt_db = 30.0
        self._xi_opt = 10.0 ** (self._xi_opt_db / 10.0)
        self._log_glr_fact = np.log(1.0 / (1.0 + self._xi_opt_db))
        self._glr_exp = self._xi_opt / (1.0 + self._xi_opt)

        self._ph1_mean = np.zeros(self._winsize) + 0.5

    def init_noise_profile(self, noise_pow):
        self._noise_pow = noise_pow

    def compute(self, frame):
        spec = np.fft.fftpack.fft(frame * self._window)
        power = np.absolute(spec) ** 2
        return self.compute_with_power_spectrum(power)

    def compute_with_power_spectrum(self, noisy_per):
        snr_post1 = noisy_per / self._noise_pow
        glr = self._prior_fact * np.exp(np.minimum(self._log_glr_fact + self._glr_exp * snr_post1, 200))
        ph1 = glr / (1.0 + glr)

        self._ph1_mean = self._alpha_ph1_mean * self._ph1_mean + (1.0 - self._alpha_ph1_mean) * ph1
        idxs = self._ph1_mean > 0.99
        ph1[idxs] = np.minimum(ph1[idxs], 0.99)
        estimate = ph1 * self._noise_pow + (1.0 - ph1) * noisy_per
        self._noise_pow = self._alpha_psd * self._noise_pow + (1.0 - self._alpha_psd) * estimate
        return self._noise_pow



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages
version = "0.1.9.5"
README = os.path.join(os.path.dirname(__file__), "README")
long_description = open(README).read() + '\n\n'
setup(name="pyssp",
      version=version,
      description=('python speech signal processing library for education'),
      long_description=long_description,
      classifiers=["Programming Language :: Python",
                   "Topic :: Scientific/Engineering",
                   "Intended Audience :: Science/Research"],
      keywords='scipy, speech processing',
      author='Shunsuke Aihara',
      author_email="s.aihara@gmail.scom",
      url='https://github.com/shunsukeaihara/pyssp',
      license="The MIT License",
      packages=find_packages(),
      namespace_packages=[],
      install_requires=["numpy", "scipy", "statsmodels", "six"]
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from statsmodels.tsa.stattools import acf
from six.moves import xrange
from pyssp.utils import calc_kurtosis


def levinson_durbin(r, order):
    """
    from http://aidiary.hatenablog.com/entry/20120415/1334458954
    """
    a = np.zeros(order + 1)
    e = np.zeros(order + 1)

    # k = 1
    a[0] = 1.0
    a[1] = - r[1] / r[0]
    e[1] = r[0] + r[1] * a[1]
    lam = - r[1] / r[0]

    for k in xrange(1, order):
        # update lambda
        lam = 0.0
        for j in xrange(k + 1):
            lam -= a[j] * r[k + 1 - j]
        lam /= e[k]

        # update a with U and V
        U = [1]
        U.extend([a[i] for i in xrange(1, k + 1)])
        U.append(0)
        V = [0]
        V.extend([a[i] for i in xrange(k, 0, -1)])
        V.append(1)
        a = np.array(U) + lam * np.array(V)

        # update e
        e[k + 1] = e[k] * (1.0 - lam * lam)

    return a, e[-1]


def lpc(frame, order):
    """
    frame: windowed signal
    order: lpc order
    return from 0th to `order`th linear predictive coefficients
    """
    r = acf(frame, unbiased=False, nlags=order)
    return levinson_durbin(r, order)[0]


def lpr_kurtosis(frame, lpcorder=10):
    """
    frame: windowed signal
    return kurtosis of linear prediction residual from input signal
    """
    c = lpc(frame, lpcorder)
    coef = c[1:][::-1] * -1
    residuals = []
    for i in xrange(frame.size - 10):
        residuals.append(frame[i + 10] - np.sum(frame[i:i + 10] * coef))
    residuals = np.array(residuals)
    return calc_kurtosis(residuals)



# -*- coding: utf-8 -*-
import scipy as sp
import numpy as np


# compute smoothing parameter
def M_H(D):
    if(D == 1):
        M = 0
        H = 0
    elif(D == 2):
        M = 0.26
        H = 0.15
    elif(D == 4):
        # nodata
        M = 0.40
        H = 0.30
    elif(D == 5):
        M = 0.48
        H = 0.48
    elif(D == 6):
        # nodata
        M = 0.50
        H = 0.65
    elif(D == 8):
        M = 0.58
        H = 0.78
    elif(D == 10):
        M = 0.61
        H = 0.98
    elif(D == 12):
        M = 0.64
        H = 1.25
    elif(D == 15):
        M = 0.668
        H = 1.55
    elif(D == 18):
        # nodata
        M = 0.68
        H = 1.7
    elif(D == 20):
        M = 0.705
        H = 2.0
    elif(D == 24):
        # no data
        M = 0.730
        H = 2.15
    elif(D == 30):
        M = 0.762
        H = 2.3
    elif(D == 40):
        M = 0.8
        H = 2.52
    elif(D == 60):
        M = 0.841
        H = 2.52
    elif(D == 80):
        M = 0.865
        H = 3.25
    elif(D == 96):
        M = 0.87
        H = 3.5
    elif(D == 120):
        M = 0.89
        H = 4.0
    elif(D == 140):
        M = 0.9
        H = 4.1
    elif(D == 160):
        M = 0.91
        H = 4.1
    elif(D == 192):
        # nodata
        M = 0.92
        H = 4.2
    else:
        M = -1
        H = -1
        print("D value:", D)
        raise ValueError('D out of range')
    return M, H


class MinimumStatistics():

    def __init__(self, winsize, window, samp_rate):
        self._window = window
        self._winsize = winsize

        # initial values
        frametime = (self._winsize / 2.0) / samp_rate  # frame incremental time [sec]
        self._snrexp = (-1.0) * frametime / 0.064
        self._av = 2.12

        self._alpha_c_lambda = 0.7
        # U subwindows
        self._U = 8
        # V samples
        self._V = 12
        # D window (96 samples)
        self._D = self._U * self._V
        self._subwc = self._V - 1
        self._ibuf = 0
        self._lmin_flag_lambda = np.zeros(self._winsize)
        self._alpha_max = 0.96
        self._beta_max = 0.8

        qeqmax = 14.0  # max value of Qeq per frame
        self._qeqimin = 1 / qeqmax

        self._clear_max = 65535 * 65535
        self._actmin_lambda = np.ones(self._winsize)
        self._actmin_lambda = self._actmin_lambda * self._clear_max
        self._actmin_lambda_sub = self._actmin_lambda
        self._Pmin_u_lambda = self._actmin_lambda
        self._actbuf = np.ones((self._U, self._winsize))
        self._actbuf = self._actbuf * self._clear_max

        # for graph
        self._x_data = []
        self._y_data = []
        self._p_data = []
        self._x_data_all = []
        self._y_data_all = []
        self._p_data_all = []
        pass

    def init_noise_profile(self, NP_lambda):
        self._P_lambda = NP_lambda
        self._sn2_lambda = self._P_lambda
        self._eP_lambda = self._P_lambda
        self._eP2_lambda = self._eP_lambda ** 2
        self._Pmin_u_lambda = self._P_lambda

    def compute(self, frame, lamda):
        Y_lambda = np.fft.fftpack.fft(frame * self._window)
        power = np.absolute(Y_lambda) ** 2
        return self.compute_with_power_spectrum(power, lamda)

    def compute_with_power_spectrum(self, power_spectrum, lamda):
        # eq9
        alpha_c_lambda_tilde = 1.0 / (1.0 + (np.sum(self._P_lambda) / np.sum(power_spectrum) - 1.0) ** 2)
        # eq10
        self._alpha_c_lambda = 0.7 * self._alpha_c_lambda + 0.3 * np.maximum(alpha_c_lambda_tilde, 0.7)
        # eq11
        alpha_lambda_hat = (self._alpha_max * self._alpha_c_lambda) / (1 + (self._P_lambda / self._sn2_lambda - 1) ** 2)
        # eq12
        snr = np.sum(self._P_lambda) / np.sum(self._sn2_lambda)
        alpha_lambda_hat = np.maximum(alpha_lambda_hat, np.minimum(0.3, snr ** self._snrexp))
        # eq4 smoothed periodgram
        self._P_lambda = alpha_lambda_hat * self._P_lambda + (1.0 - alpha_lambda_hat) * power_spectrum
        # eq20
        beta_lambda = np.minimum(alpha_lambda_hat ** 2, self._beta_max)
        self._eP_lambda = beta_lambda * self._eP_lambda + (1.0 - beta_lambda) * self._P_lambda
        self._eP2_lambda = beta_lambda * self._eP2_lambda + (1.0 - beta_lambda) * (self._P_lambda ** 2)
        # eq22
        vP_lambda = self._eP2_lambda - (self._eP_lambda ** 2)
        # eq23 modification
        Qeq_lambda_inverse = np.maximum(np.minimum(vP_lambda / (2 * (self._sn2_lambda ** 2)), 0.5),
                                        self._qeqimin / (lamda + 1))

        # eq23 + 12 lines
        eQ_lambda = np.sum(Qeq_lambda_inverse) / self._winsize
        # eq23 + 11 lines
        Bc_lambda = 1.0 + self._av * np.sqrt(eQ_lambda)

        # ----------------------------------------------------------------------------------------
        # for overall window of length D
        # ----------------------------------------------------------------------------------------

        # eq16
        M, H = M_H(self._D)
        Qeq_lambda_tilde = (1.0 / Qeq_lambda_inverse - 2 * M) / (1 - M)
        # eq17
        Bmin_lambda = 1.0 + (self._D - 1) * 2.0 / Qeq_lambda_tilde

        # ----------------------------------------------------------------------------------------
        # for subwindow U of length V
        # ----------------------------------------------------------------------------------------

        # eq16
        M, H = M_H(self._V)
        Qeq_lambda_tilde_sub = (1.0 / Qeq_lambda_inverse - 2 * M) / (1 - M)
        # eq17
        Bmin_lambda_sub = 1.0 + (self._V - 1) * 2.0 / Qeq_lambda_tilde_sub

        # ----------------------------------------------------------------------------------------

        # set k_mod(k) = 0 for all frequency bin k
        k_mod = np.zeros(self._winsize)

        # calc actmin
        actmin_lambda_current = self._P_lambda * Bmin_lambda * Bc_lambda
        # if(P*Bmin*Bc < actmin)
        k_mod = np.less(actmin_lambda_current, self._actmin_lambda)
        if(k_mod.any()):
            # update new minimum of D
            self._actmin_lambda[k_mod] = actmin_lambda_current[k_mod]
            # update new minimum of V
            actmin_lambda_current_sub = self._P_lambda * Bmin_lambda_sub * Bc_lambda
            self._actmin_lambda_sub[k_mod] = actmin_lambda_current_sub[k_mod]

        if(0 < self._subwc and self._subwc < self._V - 1):
            # sample is NOT the fisrt or the last; middle of buffer allows a local minimum
            self._lmin_flag_lambda = np.minimum(self._lmin_flag_lambda + k_mod, 1)
            # calc Pmin_u_lamda == sigma_n_lamda_hat**2
            self._Pmin_u_lambda = np.minimum(self._actmin_lambda_sub, self._Pmin_u_lambda)
            # store sn2 for later use
            self._sn2_lambda = self._Pmin_u_lambda
            self._subwc = self._subwc + 1
        elif(self._subwc >= self._V - 1):
            # sample IS the last; end of buffer lets finishing subwindow process and a buffer switch
            self._lmin_flag_lambda = np.maximum(self._lmin_flag_lambda - k_mod, 0)
            # store actmin_lamnda, note actbuf is NOT cyclic!
            self._ibuf = np.mod(self._ibuf, self._U)
            self._actbuf[self._ibuf] = self._actmin_lambda
            self._ibuf = self._ibuf + 1
            # find Pmin_u, the minimum of the last U stored value of actmin
            self._Pmin_u_lambda = self._actbuf.min(axis=0)

            # calc noise_slope_max
            if(eQ_lambda < 0.03):
                noise_slope_max = 8.0
            elif(eQ_lambda < 0.05):
                noise_slope_max = 4.0
            elif(eQ_lambda < 0.06):
                noise_slope_max = 2.0
            else:
                noise_slope_max = 1.2
            # replace all previously stored values of actmin by actminsub
            lmin = self._lmin_flag_lambda * np.less(self._actmin_lambda_sub,
                                                    noise_slope_max * self._Pmin_u_lambda) *\
                  np.less(self._Pmin_u_lambda, self._actmin_lambda_sub)
            lmin = np.int16(lmin)
            if(lmin.any()):
                self._Pmin_u_lambda[lmin] = self._actmin_lambda_sub[lmin]
                r = np.ones((self._U, self._winsize))
                rp = r * self._Pmin_u_lambda
                self._actbuf[:, lmin] = rp[:, lmin]

            # update flags
            self._lmin_flag_lambda = np.zeros(self._winsize)
            self._actmin_lambda = np.ones(self._winsize) * self._clear_max
            self._subwc = 0
        else:
            self._subwc = self._subwc + 1

        x = self._sn2_lambda
        # stderror; paper [2], Fig 15; not use
        # qisq = np.sqrt(Qeq_lambda_inverse)
        # xs = x*np.sqrt(0.266*(self._D+100*qisq)*qisq/\
        # (1+0.005*self._D+6.0/self._D)/(0.5*(1.0/Qeq_lambda_inverse)+self._D-1))
        self._x_data.append(x)

        self._x_data_all.append(10 * np.log10(np.sum(np.absolute(x)) / self._winsize))
        self._p_data_all.append(10 * np.log10(np.sum(self._P_lambda) / self._winsize))
        self._y_data.append(power_spectrum)
        self._y_data_all.append(10 * np.log10(np.sum(power_spectrum) / self._winsize))
        return x

    def show_debug_result(self):
        import matplotlib.pyplot as plt
        fig = plt.figure()

        ax = fig.add_subplot(111)

        ax.grid('on', 'major', color='#666666', linestyle='-')

        ax.plot(self._y_data_all)
        ax.plot(self._p_data_all, linewidth=3.0)
        ax.plot(self._x_data_all, linewidth=3.0)
        ax.legend(("Periodgram", "Smoothed Periodgram", "Estimated Noise"))  # ,"True Noise"))
        plt.show()



# -*- coding: utf-8 -*-
import numpy as np


class RecursiveAveragingEstimator:
    # http://www.sciencedirect.com/science/article/pii/S0167639305002001
    def __init__(self, winsize, window, samp_rate,
                 nu=0.7, B=0.8, gamma=0.998, alpha_p=0.2, alpha_d=0.85):

        def calc_delta():
            delta = np.zeros(winsize, dtype=np.float64)
            return delta

        self._window = window
        self._winsize = winsize
        # initial values
        frametime = (self._winsize / 2.0) / samp_rate  # frame incremental time [sec]
        self._snrexp = (-1.0) * frametime / 0.064
        self._av = 2.12

        self._nu = nu
        self._B = B
        self._gamma = gamma
        self._alpha_p = alpha_p
        self._alpha_d = alpha_d

        self._delta = calc_delta()

        self._nplist = []
        self._nsplist = []
        self._snsplist = []

    def init_noise_profile(self, NP):
        self._lastD = NP
        self._lastP = NP
        self._lastPmin = NP
        self._lastp = np.zeros(self._winsize, dtype=np.float64)

    def compute(self, frame, lamda):
        Y = np.fft.fftpack.fft(frame * self._window)
        Ysq = np.absolute(Y) ** 2.0
        return self.compute_with_power_spectrum(Ysq, lamda)

    def compute_with_power_spectrum(self, power_spectrum, lamda):
        Ysq = power_spectrum
        # Compute smooth speech power spectrum
        P = self._nu * self._lastP + (1.0 - self._nu) * Ysq  # eq 2

        # Find the local minimum of noisy speech
        Pmin = np.zeros(self._winsize, dtype=np.float64)
        minidx = np.less_equal(P, self._lastPmin)
        largeidx = np.greater(P, self._lastPmin)
        Pmin[minidx] = P[minidx]
        Pmin[largeidx] = self._gamma * self._lastPmin[largeidx] + \
                        ((1.0 - self._gamma) / (1.0 - self._B)) *\
                        (P[largeidx] - self._B * self._lastP[largeidx])  # eq3
        # Compute ratio of smoothed speech p
        Sr = P / Pmin  # eq4

        # Calculate speech presence probability using first-order recursion
        I = np.array(np.greater(self._delta, Sr), dtype=np.float64)  # eq5
        p = self._alpha_p * self._lastp + (1.0 - self._alpha_p) * I  # eq6

        # Compute time-frequency dependent smoothing factors
        alpha_s = self._alpha_d + (1.0 - self._alpha_d) * p  # eq7

        # Update noise estimate using time-frequency dependent smoothing factors
        D = alpha_s * self._lastD + (1.0 - alpha_s) * Ysq  # eq8

        # calculate debug info
        noise_periodgram = 10 * np.log10(np.sum(np.absolute(D)) / self._winsize)
        noisy_speech_periodgram = 10 * np.log10(np.sum(Ysq) / self._winsize)
        smooth_noisy_speech_periodgram = 10 * np.log10(np.sum(P) / self._winsize)
        self._nplist.append(noise_periodgram)
        self._nsplist.append(noisy_speech_periodgram)
        self._snsplist.append(smooth_noisy_speech_periodgram)
        # update
        self._lastP = P
        self._lastPmin = Pmin
        self._lastp = p
        self._lastD = D
        return D

    def show_debug_result(self):
        import matplotlib.pyplot as plt
        fig = plt.figure()

        ax = fig.add_subplot(111)

        ax.grid('on', 'major', color='#666666', linestyle='-')

        ax.plot(self._nsplist)
        ax.plot(self._snsplist, linewidth=3.0)
        ax.plot(self._nplist, linewidth=3.0)
        ax.legend(("Periodgram", "Smoothed Periodgram", "Estimated Noise"))  # ,"True Noise"))
        plt.show()



# -*- coding: utf-8 -*-
import scipy as sp
import numpy as np


def calc_nullhypotes(pa, pp, alpha=1.0):
    return (1.0 / (np.sqrt(np.pi * 2) * alpha * pa)) * np.exp((-1.0 / 2 * (alpha ** 2)) * (pp / pa) ** 2)


def calc_hypotes(pa, pp, beta=1.0):
    return (1.0 / (np.sqrt(np.pi * 2) * beta * pp)) * np.exp((-1.0 / 2 * (beta ** 2)) * (pa / pp) ** 2)


class PAR:
    def __init__(self, winsize, window, alpha=1.0, beta=1.0):
        self._winsize = winsize
        self._window = window
        self._alpha = alpha
        self._beta = beta
        self._eta = (2.0 * (window ** 2).sum()) / window.sum() ** 2

    def calc_par(self, frame):
        power = sp.absolute(sp.fft(frame * self._window)) ** 2
        avg_pow = power[:int(self._winsize / 2)].sum() / (self._winsize / 2)
        smax = -np.inf
        lenmax = 0
        for i in range(2, int(self._winsize / 10)):
            # searching f0 with maximizing estimated power of periodic component
            idx = list(range(i, int(self._winsize / 2), i + 1))
            score = power[:int(self._winsize / 2)][idx].sum() - (len(idx) * avg_pow)
            if score > smax:
                smax = score
                lenmax = len(idx)
        pp = (smax / (1.0 - self._eta * lenmax)) * self._eta
        pa = avg_pow - pp
        return calc_hypotes(pa, pp, beta=self._beta) / calc_nullhypotes(pa, pp, alpha=self._alpha)

    
